#! /bin/sh

# 30 January 20092

# This is the official template for pipetool regression test scripts.
# In addition to supporting the "SHORTTEST" option, this script also
# allows the user to run individual subtests from the command line.
# The script will accept a series of test identifiers, generally of
# the form "test1" "test2" ... which are to be run.


# acis_streak_map.t
# test script for acis_streak_map


# syntax:
# acis_streak_map.t [<testid> ... ]
 



######################################################################
# subroutine
# error_exit <message>
# Fatal error exit

error_exit()
{
  echo "$1" | tee -a $LOGFILE
  echo "${toolname} : FAIL" | tee -a $LOGFILE
  exit 1
}

######################################################################
# subroutine
# keyfilter infile outfile
# filters out CHECKSUM, Dataset, CREATOR, HISTORY, DATASUM, 
#             ASCDSVER, HISTNUM, and DATE
# To filter additional keywords, add s/KEYWORD/Dataset/g; for each.

keyfilter()
{
  cat $1 | sed -e 's/CHECKSUM/Dataset/g;s/COMMENT/Dataset/g;
  s/DATE/Dataset/g;s/CREATOR/Dataset/g;s/HISTORY/Dataset/g;
  s/DATASUM/Dataset/g;s/ASCDSVER/Dataset/g;s/HISTNUM/Dataset/g' | \
  grep -v Dataset > $2
  zerotest $2
}

######################################################################
# subroutine
# find_tool <toolname>
# checks that tool exists and is runnable

find_tool()
{
  s1=`type $1`
  s2=`echo $s1 | awk -F" " '{ print $3}'`
  if test -x $s2 ; then
    :
  else
    error_exit "tool $1 not found"
  fi
}

######################################################################
# subroutine
# zerotest <file> 
# Makes sure that file is not 0 length.
# Use this to protect yourself against empty files  (which will 
# 'diff' without error).  This can happen when the input file to
# cat $infile | do_something >> $outfile
# is missing.  This is used by keyfilter(), above.

zerotest()
{
 if test -s $1 ;
 then
   :
 else
   echo "ERROR: file $1 is of zero length" >> $LOGFILE
   #  Indicate failure, but do not exit.
   mismatch=0
 fi
}


######################################################################
# Initialization

toolname="acis_streak_map"

# set up list of tests
alltests="test1 test2 test3 test637"

# "short" test to run
shortlist="test2"

# compute date string for log file
DT=`date +'%d%b%Y_%T'`

# check for directory environment variables
if test "x${TESTIN}" = "x" -o "x${TESTOUT}" = "x" -o "x${TESTSAV}" = "x" \
   -o "x${TESTLOG}" = "x" ; then
  error_exit "one or more of TESTIN/TESTOUT/TESTSAV/TESTLOG not defined" 
fi

# convenience definitions
OUTDIR=$TESTOUT/$toolname
SAVDIR=$TESTSAV/$toolname
INDIR=$TESTIN/$toolname
LOGDIR=$TESTLOG/$toolname

# set up log file name
LOGFILE=$LOGDIR/${toolname}_log.$DT

#get rid of old logs
rm -f $LOGDIR/${toolname}_log.*



# Any tests specified on command line?
if test $# -gt 0; then
  # yes, do those tests
  testlist=$*
else
  # No, see if we are to do "short" test
  if test "x$SHORTTEST" = "x" ; then
    # No, do everything
    testlist=$alltests
  else
    # yes, do short test
    testlist=$shortlist
  fi
fi


# Make sure we have a log directory
if test -d $LOGDIR ; then
 :
else
  mkdir -p $LOGDIR 
  if test $? -ne 0 ; then
    error_exit ""
  fi
fi


# Make sure we have an output directory
if test -d $OUTDIR ; then
 :
else
  mkdir -p $OUTDIR >> $LOGFILE 2>&1
  if test $? -ne 0 ; then
    error_exit "can't create output directory $OUTDIR"
  fi
fi


# check for tools
# if a utility is used in the form "utility <args> > outfile", and 'utility'
# cannot be run, 'outfile' will still be created.  If utility is used on 
# both the output and reference files of a tool the resultant utility output 
# files will both exist and be empty, and will pass a diff.

find_tool dmlist
find_tool dmimgcalc



# announce ourselves
echo ""
echo "${toolname} regression" | tee $LOGFILE
echo ""

# All parameters except verbose should be set anyway, but clear them
# to be safe.
bose=`pget $toolname verbose`
punlearn $toolname
pset $toolname verbose=$bose

script_succeeded=0

######################################################################
# Begin per-test loop

for testid in $testlist
do

  # delete old outputs
  rm -f $OUTDIR/${testid}*

  # Set up file names
  outroot=$OUTDIR/${testid}_bkg
  outregfile=$OUTDIR/${testid}_streakreg3.fits
  savroot=$SAVDIR/${testid}_bkg
  savregfile=$SAVDIR/${testid}_streakreg3.fits

  echo "" | tee -a $LOGFILE
  echo "running $testid" | tee -a $LOGFILE
  echo "" | tee -a $LOGFILE
  ####################################################################
  # run the tool
  case ${testid} in

    test1 ) #...Define input files, params
            chipid="7"
            evtfile="${INDIR}/${testid}_evt3_ccdid7.fits[energy=500:7000]"
            fovfile="${INDIR}/${testid}_fov3.fits"
            test2_string="acis_streak_map infile=$evtfile \
                     fovfile=$fovfile bkgroot=$outroot  \
                     regfile=$outregfile ssigma=-1 clobber=\"yes\""
            echo ${test2_string} | tee -a  $LOGFILE
            eval ${test2_string} | tee -a  $LOGFILE  2>&1
            ;;

    test2 ) #...Define input files, params 
            chipid="0 1 2 3" 
	    evtfile="${INDIR}/${testid}_evt3.fits[energy=500:7000]" 
	    fovfile="${INDIR}/${testid}_fov3.fits" 
	    test2_string="acis_streak_map infile=$evtfile \
	             fovfile=$fovfile bkgroot=$outroot  \
	             regfile=$outregfile ssigma=-1 clobber=\"yes\""
	    echo ${test2_string} | tee -a  $LOGFILE 
	    eval ${test2_string} | tee -a  $LOGFILE  2>&1
            ;;

    test3 ) #...Define input files, params 
	    #...This test checks ssigma > 0, ie, break streaks around srcs
            chipid="2 3 5 6 7 8" 
	    evtfile="${INDIR}/${testid}_evt3.fits[energy=500:7000]" 
	    fovfile="${INDIR}/${testid}_fov3.fits" 
	    test3_string="acis_streak_map infile=$evtfile \
	             fovfile=$fovfile bkgroot=$outroot  \
	             regfile=$outregfile ssigma=6 clobber=\"yes\""
	    echo ${test3_string} | tee -a  $LOGFILE 
	    eval ${test3_string} | tee -a  $LOGFILE  2>&1
            ;;

    test637 ) #...Define input files, params 
            chipid="0 1 2 3 6 7" 
	    evtfile="${INDIR}/test637_evt2.fits[energy=500:7000]" 
	    fovfile="${INDIR}/test637_fov1.fits" 
	    test2_string="acis_streak_map ssigma=-1 infile=$evtfile \
	             fovfile=$fovfile bkgroot=$outroot  \
	             regfile=$outregfile clobber=\"yes\""
	    echo ${test2_string} | tee -a  $LOGFILE 
	    eval ${test2_string} | tee -a  $LOGFILE  2>&1
            ;;
  esac

  ####################################################################
  # check the outputs

  # Init per-test error flag
  mismatch=1

  ####################################################################
  # FITS table

  # new output
  #dmlist $outfile opt=header,data,clean rows=700:1400 > \
  # $OUTDIR/${testid}.dmp1 2>>$LOGFILE
  #keyfilter $OUTDIR/${testid}.dmp1 $OUTDIR/${testid}.dmp2 \
  # 2>>$LOGFILE
  
  # reference output
  #dmlist $savfile opt=header,data,clean rows=700:1400 > \
  # $OUTDIR/${testid}.dmp1_std 2>>$LOGFILE
  #keyfilter $OUTDIR/${testid}.dmp1_std $OUTDIR/${testid}.dmp2_std \
  # 2>>$LOGFILE
  
  # compare
  #diff $OUTDIR/${testid}.dmp2 $OUTDIR/${testid}.dmp2_std > \
  # /dev/null 2>>$LOGFILE

  # 1. check per chip streakmap files
  #  eval chipid='$'${testid}_chips
  for chip in $chipid
  do
    dmdiff ${outroot}${chip}.fits ${savroot}${chip}.fits tol=$SAVDIR/tolerance > /dev/null 2>>$LOGFILE
    if  test $? -ne 0 ; then
      echo "ERROR: MISMATCH in ${outroot}${chip}.fits" >> $LOGFILE
      mismatch=0
    fi
  done  

  dmdiff ${outregfile}  ${savregfile} > /dev/null 2>>$LOGFILE
  if  test $? -ne 0 ; then
     echo "ERROR: MISMATCH in ${outregfile}" >> $LOGFILE
     mismatch=0
  fi

  ####################################################################
  # Did we get an error?
  if test $mismatch -eq 0 ; then
    # Yes
    echo "${testid} NOT-OK"
    script_succeeded=1
  else
    # No
    echo "${testid} OK"
  fi

done
# end per-test loop
######################################################################


######################################################################
# report results

# blank line
echo ""

if test $script_succeeded -eq 0; then
    echo "${toolname} : PASS" | tee -a $LOGFILE
else
    echo "${toolname} : FAIL" | tee -a $LOGFILE
fi

echo "log file in ${LOGFILE}"


exit $script_succeeded
