#! /bin/sh

# Nov. 2012

# This is the official template for pipetool regression test scripts.
# In addition to supporting the "SHORTTEST" option, this script also
# allows the user to run individual subtests from the command line.
# The script will accept a series of test identifiers, generally of
# the form "test1" "test2" ... which are to be run.

# Portions of the script which must be customized are marked with "!!",
# below.

# The complete list of tests must be placed in "alltests", !!3, below.
# The test[s] for the SHORTTEST must be placed in "shortlist", !!4 below.

# !!1
# tgdetect2.t
# test script for tgdetect2

# !!2
# syntax:
# tgdetect2.t [<testid> ... ]
 
######################################################################
# subroutine
# error_exit <message>
# Fatal error exit

error_exit()
{
  echo ""   | tee -a $LOGFILE
  echo "$1" | tee -a $LOGFILE
  echo ""   | tee -a $LOGFILE
  echo "${toolname} : FAIL" | tee -a $LOGFILE
  exit 1
}

######################################################################
# subroutine
# keyfilter infile outfile
# filters out CHECKSUM, Dataset, CREATOR, HISTORY, DATASUM, 
#             ASCDSVER, HISTNUM, and DATE
# To filter additional keywords, add s/KEYWORD/Dataset/g; for each.

keyfilter()
{
  cat $1 | sed -e 's/ROLL_RMS/Dataset/g;s/COMMENT/Dataset/g;
  s/CHECKSUM/Dataset/g; s/DATE/Dataset/g;s/CREATOR/Dataset/g;
  s/HISTORY/Dataset/g; s/DATASUM/Dataset/g;s/ASCDSVER/Dataset/g;
  s/HISTNUM/Dataset/g' | \
  grep -v Dataset > $2
  zerotest $2
}

######################################################################
# subroutine
# find_tool <toolname>
# checks that tool exists and is runnable

find_tool()
{
  s1=`type $1`
  s2=`echo $s1 | awk -F" " '{ print $3}'`
  if test -x $s2 ; then
    :
  else
    error_exit "tool $1 not found"
  fi
}

######################################################################
# subroutine
# zerotest <file> 
# Makes sure that file is not 0 length.
# Use this to protect yourself against empty files  (which will 
# 'diff' without error).  This can happen when the input file to
# cat $infile | do_something >> $outfile
# is missing.  This is used by keyfilter(), above.

zerotest()
{
 if test -s $1 ;
 then
   :
 else
   echo "ERROR: file $1 is of zero length" >> $LOGFILE
   #  Indicate failure, but do not exit.
   mismatch=0
 fi
}

######################################################################
# Initialization

# !!3
toolname="tgdetect2"

# set up list of tests
# !!4
alltests="test1 test2 test3 test4 filter1 filter2 filter3 filter4 filter5 tg_findzo" 

# "short" test to run
# !!5
shortlist="test1"


# compute date string for log file
DT=`date +'%d%b%Y_%T'`

# convenience definitions
OUTDIR=$TESTOUT/$toolname
SAVDIR=$TESTSAV/$toolname
INDIR=$TESTIN/$toolname
LOGDIR=$TESTLOG/$toolname

# set up log file name
LOGFILE=$LOGDIR/${toolname}_log.$DT

#get rid of old logs
rm -f $LOGDIR/${toolname}_log.*

# Any tests specified on command line?
if test $# -gt 0; then
  # yes, do those tests
  testlist=$*
else
  # No, see if we are to do "short" test
  if test "x$SHORTTEST" = "x" ; then
    # No, do everything
    testlist=$alltests
  else
    # yes, do short test
    testlist=$shortlist
  fi
fi

# Make sure we have a log directory
if test -d $LOGDIR ; then
 :
else
  mkdir -p $LOGDIR 
  if test $? -ne 0 ; then
    error_exit ""
  fi
fi

# Make sure we have an output directory
if test -d $OUTDIR ; then
 :
else
  mkdir -p $OUTDIR >> $LOGFILE 2>&1
  if test $? -ne 0 ; then
    error_exit "can't create output directory $OUTDIR"
  fi
fi

# Make sure we have an input directory
if test -d $INDIR ; then
:
else
  if test $? -ne 0 ; then
    error_exit "can't find input directory : $INDIR"
  fi
fi

# Make sure we have a save directory
if test -d $SAVDIR ; then
:
else
  if test $? -ne 0 ; then
    error_exit "can't find save directory : $SAVDIR"
  fi
fi

# check for directory environment variables
if test "x${TESTIN}" = "x" -o "x${TESTOUT}" = "x" -o "x${TESTSAV}" = "x" \
   -o "x${TESTLOG}" = "x" ; then
  error_exit "one or more of TESTIN/TESTOUT/TESTSAV/TESTLOG not defined" 
fi


# check for tools
# if a utility is used in the form "utility <args> > outfile", and 'utility'
# cannot be run, 'outfile' will still be created.  If utility is used on 
# both the output and reference files of a tool the resultant utility output 
# files will both exist and be empty, and will pass a diff.

find_tool dmdiff

# announce ourselves
echo ""
echo "${toolname} regression" | tee $LOGFILE
echo "" | tee -a $LOGFILE

# All parameters except verbose should be set anyway, but clear them
# to be safe.
bose=`pget $toolname verbose`

script_succeeded=0

######################################################################
# Begin per-test loop
for testid in $testlist
do
  ####################################################################
  # Init per-test error flag
  #

  punlearn $toolname
  pset $toolname verbose=$bose

  mismatch=1    

  # Set up file names
  outfile=$OUTDIR/${testid}_src1a.fits
  savfile=$SAVDIR/${testid}_src1a.fits

  # delete old outputs
  rm -f ${outfile}

  echo "running $testid" | tee -a $LOGFILE
  echo "" | tee -a $LOGFILE
  ####################################################################
  # run the tool
  case ${testid} in
    # !!6
    # output from tg_findzo;
    test1 ) test1_string="tgdetect2 infile=${INDIR}/2739_evt1.fits \
               outfile=${outfile} zo_pos_x=default zo_pos_y=default \
               verb=0 mode=h "
            ;;
    # output from tgdetect;
    test2 ) test1_string="tgdetect2 infile=${INDIR}/7409_evt1.fits \
               outfile=${outfile} zo_pos_x=default zo_pos_y=default \
               verb=0 mode=h "
            ;;
    # grating=none; output from tgdetect;
    test3 ) test1_string="tgdetect2 infile=${INDIR}/2739_evt1_b.fits \
               outfile=${outfile} zo_pos_x=default zo_pos_y=default \
               verb=0 mode=h "
            ;;
    # bug 13729_I : no evt data; run tgdetect;
    test4 ) test1_string="tgdetect2 infile=${INDIR}/862_all_flt_evt1.fits \
               outfile=${outfile} zo_pos_x=default zo_pos_y=default \
               verb=0 mode=h "
            ;;
    # bug 13729_II :  if tg_findzo fail to run,  continue tgdetect.
    # test5 ) test1_string="tgdetect2 infile=${INDIR}/2739_evt1_c.fits \
    #          outfile=${outfile} zo_pos_x=default zo_pos_y=default \
    #          verb=0 mode=h "
    #       ;;
    # bug 13774 : tgdet_1 (bug 5786) can't accept infile w/ dmfilter
    filter1 ) test1_string="tgdetect2 infile='${INDIR}/7490_evt1.fits[status=0,grade=0,2,3,4,6]' \
               outfile=${outfile} zo_pos_x=default zo_pos_y=default \
               verb=0 mode=h "
            ;;
    # bug 13774 : can't accept infile w/ dmfilter
    filter2 ) test1_string="tgdetect2 infile='${INDIR}/7490_evt1.fits[grade<20]' \
               outfile=${outfile} zo_pos_x=default zo_pos_y=default \
               verb=0 mode=h "
            ;;
    # bug 13774 : can't accept infile w/ dmfilter
    filter3 ) test1_string="tgdetect2 infile='${INDIR}/7490_evt1.fits[exclude grade=7]' \
               outfile=${outfile} zo_pos_x=default zo_pos_y=default \
               verb=0 mode=h "
            ;;
    # bug 13774 : can't accept infile w/ dmfilter
    filter4 ) test1_string="tgdetect2 infile='${INDIR}/7490_evt1.fits[cols *]' \
               outfile=${outfile} zo_pos_x=default zo_pos_y=default \
               verb=0 mode=h "
            ;;
    # bug 13774 : can't accept infile w/ dmfilter
    filter5 ) test1_string="tgdetect2 infile='${INDIR}/7490_evt1.fits[events][status=0][sky=field()]' \
               outfile=${outfile} zo_pos_x=default zo_pos_y=default \
               verb=0 mode=h "
            ;;
    # SAT-96
    tg_findzo ) test1_string="tgdetect2 infile=${INDIR}/all_flt_acisf00660_000N001_evt1.fits outfile=${outfile} zo_pos_x=0 zo_pos_y=0 unlearn_tgdetect=no unlearn_tg_findzo=no temproot=${INDIR}/acisf00660_000N001 keepexit=yes verb=0"
            ;;

  esac

  echo ${test1_string} | tee -a  $LOGFILE
  eval ${test1_string} >/dev/null 2>>$LOGFILE

  ####################################################################
  # if the tool failed to run, set mismatch to 0.
  #
  if test $? -ne 0; then
     echo "" | tee -a $LOGFILE
     echo "$toolname failed to run" | tee -a $LOGFILE
     echo "" | tee -a $LOGFILE
     mismatch=0           
  fi

  ####################################################################
  # check the outputs

  # if different tests need different kinds of comparisons, use a 
  #  case ${testid} in...  here

  ####################################################################
  # FITS table    (duplicate for as many tables per test as needed)

  # compare
  dmdiff $outfile $savfile tol=$SAVDIR/tolerance subspace- >/dev/null 2>>$LOGFILE
  if  test $? -ne 0 ; then
    echo "ERROR: MISMATCH in $outfile" >> $LOGFILE
    mismatch=0
  fi

  ####################################################################
  # Did we get an error?
  if test $mismatch -eq 0 ; then
    # Yes
    echo "${testid} NOT-OK"
    script_succeeded=1
  else
    # No
    echo "${testid} OK"
  fi

  echo "" | tee -a $LOGFILE

done
# end per-test loop
######################################################################


######################################################################
# report results

# blank line
echo "" | tee -a $LOGFILE

if test $script_succeeded -eq 0; then
    echo "${toolname} : PASS" | tee -a $LOGFILE
else
    echo "${toolname} : FAIL" | tee -a $LOGFILE
fi

echo ""
echo "log file in ${LOGFILE}"
echo ""

exit $script_succeeded
